//Hangeul Typing Behavior Analyzer
//created by Kim Yongmook ( http://moogi.new21.org )
//Originically created on March 7, 2000
//Modified on April 21, 2005

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
 
#define FALSE 0
 
//Member: 0->Shift?, 1->Finger, 2->Height in keyboard
char typeinfo[128][3];
 
#define sett(n,a,b,c) typeinfo[n][0]=a, typeinfo[n][1]=b, typeinfo[n][2]=c;
#define is_special(ch1) \
	((int)strchr("0xd40xd90xda0xdb0xdc0xdd0xde",ch1))
#define is_hanja(ch1,ch2) \
	( ch1>=0xE0 && ch1<=0xF9 && ((ch2>=0x31 && ch2<=0x7E) || (ch2>0x91 && ch2<=0xFE)) )
 
void init_type_info(int n)
{
	int i;
	sett('a',0,-5,3); sett('s',0,-4,3); sett('d',0,-3,3); sett('f',0,-2,3); sett('g',0,-2,3);
	sett('h',0,2,3); sett('j',0,2,3); sett('k',0,3,3); sett('l',0,4,3);
 
	sett('q',0,-5,2); sett('w',0,-4,2); sett('e',0,-3,2); sett('r',0,-2,2); sett('t',0,-2,2);
	sett('y',0,2,2); sett('u',0,2,2); sett('i',0,3,2); sett('o',0,4,2); sett('p',0,5,2);
 
	sett('z',0,-5,4); sett('x',0,-4,4); sett('c',0,-3,4); sett('v',0,-2,4);
	sett('b',0,(n==2 ? 2:-2),4);
	sett('n',0,2,4); sett('m',0,2,4);
	for(i='A';i<='Z';i++) sett(i,1,typeinfo[i+32][1],typeinfo[i+32][2]);
 
	sett('1',0,-5,1); sett('2',0,-4,1); sett('3',0,-3,1); sett('4',0,-2,1); sett('5',0,-2,1);
	sett('6',0,-2,1); sett('7',0,2,1); sett('8',0,3,1); sett('9',0,4,1); sett('0',0,5,1);
 
	sett('!',1,-5,1); sett('@',1,-4,1); sett('#',1,-3,1); sett('$',1,-2,1); sett('%',1,-2,1);
	sett('^',1,-2,1); sett('&',1,2,1); sett('*',1,3,1); sett('(',1,4,1); sett(')',1,5,1);
 
	sett('[',0,5,2); sett(']',0,5,2);
	sett(';',0,5,3); sett('/',0,5,4); sett('\'',0,5,3);
	sett('-',0,5,1); sett('=',0,5,1); sett('\\',0,5,1);
}
 
void parse(char *keyseq)
{
	int i,j, shift=FALSE, handst=0, fingerst=0, kheight=-1, sgk, abk;
	int errshift=0, errnear=0, errnear2=0;
	int finger[6]={0,}, kleft=0, kright=0;
	//puts(keyseq);
	for(i=0,j=strlen(keyseq);i<j;i++) {
		if(typeinfo[keyseq[i]][0]!=shift && shift==0) {
			//printf("Pos %d(%c): ",i, keyseq[i]);
			//puts( (shift=typeinfo[keyseq[i]][0]) ? "  Hold shift key": "  Up shift key");
			errshift++; finger[5]++;
		}
		sgk=typeinfo[keyseq[i]][1]==0 ? 0: typeinfo[keyseq[i]][1]>0 ? 1:-1;
		abk=typeinfo[keyseq[i]][1]>0 ? typeinfo[keyseq[i]][1]:-typeinfo[keyseq[i]][1];
		if(abk!=typeinfo[keyseq[i]][1]) kleft++; else kright++;
		if(handst==sgk && abs(fingerst-abk)<=1 && abs(typeinfo[keyseq[i]][2]-kheight)<=1) {
			//printf("Pos %d(%c): ",i, keyseq[i]);
			//puts("  Type with near finger");
			errnear++;
		}
		if(handst==sgk && fingerst==abk) errnear2++;
		handst=sgk; fingerst=abk; kheight=typeinfo[keyseq[i]][2];
		shift=typeinfo[keyseq[i]][0];
		finger[abk]++;
	}
	puts("Result:");
	printf("Total key presses: %d",j);
	if(errshift) printf("(%d, including shift)\n",j+errshift); else puts("");
	printf("Total shift presses: %d\n",errshift);
	printf("Total near-key presses: %d (%d)\n",errnear, errnear2);
	finger[1]=finger[2]+finger[3]+finger[4]+finger[5];
	for(i=2;i<=5;i++) printf("%d:%d:%2.1f%%	",i,finger[i],100.0*finger[i]/finger[1] ); puts("");
}
 
void parse_han(unsigned char *h,int *flag,int *cho,int *jung,int *jong)
{
	if(flag) *flag=h[0] & 128;
	if(cho)  *cho =(h[0] & (64+32+16+8+4))>>2;
	if(jung) *jung=((h[0] & (2+1))<<3)|((h[1] & (128+64+32))>>5);
	if(jong) *jong=(h[1] & (1+2+4+8+16));
}
 
/* ѱ ڼ ڵ
 ʼ: NL(1) (2) (3) (4) (5) (6) (7) (8) (9) (10) (11) (12)
	   (13) (14) (15) (16) (17) (18) (19) (20)
*/
char cho2[21][3]={
	"", "",
	"r", "R", "s", "e", "E", "f", "a", "q", "Q", "t", "T", // to 
	"d", "w", "W", "c", "z", "x", "v", "g"
};
char cho3[21][3]={
	"", "",
	"k", "kk", "h", "u", "uu", "y", "i", ";", ";;", "n", "nn", // to 
	"j", "l", "ll", "o", "0", "'", "p", "m"
};
/*
 ߼: NL(2) (3) (4) (5) (6) (7) (10) (11) (12) (13) (14)
	   (15) (18) (19) (20) (21) (22) (23) (26) (27) (28) (29)
*/
char jung2[30][3]={
	"", "","",
	"k", "o", "i", "O", "j", "", "",
	"p", "u", "P", "h", "hk", // to 
	"ho", "", "", "hl",
	"y", "n", "nj", "np", "nl", "", "",
	"b", "m", "ml", "l"
};
char jung3[30][3]={
	"", "","",
	"f", "r", "6", "R", "t", "", "",
	"c", "e", "7", "v", "/f", // to 
	"/r", "", "", "/d",
	"4", "b", "9t", "9c", "9d", "", "",
	"5", "g", "8", "d"
};
char jung3f[30][3]={
	"", "","",
	"f", "r", "6", "G", "t", "", "",
	"c", "e", "7", "v", "/f", // to 
	"/r", "", "", "/d",
	"4", "b", "9t", "9c", "9d", "", "",
	"5", "g", "8", "d"
};
char jung3n[30][3]={
	"", "","",
	"f", "r", "6", "0", "t", "", "",
	"c", "e", "7", "v", "vf", // to 
	"vr", "", "", "vd",
	"4", "b", "bt", "bc", "bd", "", "",
	"5", "g", "8", "d"
};
/*
: NL(1) (2) (3) (4) (5) (6) (7) (8) (9) (10) (11) (12)
	   (13) (14) (15) (16) (17) (19) (20) (21) (22) (23) (24)
	   (25) (26) (27) (28) (29)
*/
char jong2[30][3]={
	"","",
	"r","R","rt","s","sw","sg","e","f","fr","fa","fq",
	"ft","fx","fv","fg","a", "",
	"q","qt","t","T","d","w","c","z","x","v","g"
};
char jong3[30][3]={
	"","",
	"x","F","xq","s","s!","S","A","w","D","C","w3",
	"wq","wW","wQ","V","z","", 
	"3","X","q","2","a","!","Z","E","W","Q","1"
};
char jong3f[30][3]={
	"","",
	"x","!","V","s","E","S","A","w","@","F","D",
	"T","%","$","R","z","", 
	"3","X","q","2","a","#","Z","C","W","Q","1"
};
char jong3n[30][3]={
	"","",
	"x","xx","xq","s","s-","s1","/","w","wx","wz","w3", 
	"wq","w[","w]","w1","z","",
	"3","3q","q","2","a","-","=","\\","[","]","1"
};
 
void main(int argc, char **argv)
{
	FILE *fp; unsigned char *buff; int f=0;
	int i,j, cho,jung,jong, hanguls=0;
 
	char *key2, *key3, *key3f, *key3n;
	int pos2,pos3,pos3f,pos3n;
 
	fp=fopen((argc==2 ? argv[1]:"sample.txt"),"rt");
	if(fp==NULL) {puts("Cannot open file."); exit(1);}
	
	fseek(fp,0,SEEK_END); f=ftell(fp);
	printf("Reading %d bytes\n",f); fseek(fp,0,SEEK_SET);
	buff=(unsigned char *)calloc(f,1); fread(buff,f,1,fp);
 
	key2=(char *)calloc(f*2,1); key3=(char *)calloc(f*2,1);
	key3f=(char *)calloc(f*2,1); key3n=(char *)calloc(f*2,1);
	pos2=pos3=pos3f=pos3n=0;
 
	fclose(fp);
	for(i=0,j=strlen((char *)buff);i<j;i++)
		if(buff[i]>=128) {
			if(is_hanja(buff[i],buff[i+1])) {i++;continue;}
			if(is_special(buff[i])) {i++;continue;}
			parse_han(buff+i, NULL, &cho, &jung, &jong); hanguls++;
			strcpy(key2+pos2, cho2[cho]); pos2+=strlen(cho2[cho]);
			strcpy(key2+pos2, jung2[jung]); pos2+=strlen(jung2[jung]);
			strcpy(key2+pos2, jong2[jong]); pos2+=strlen(jong2[jong]);
 
			strcpy(key3+pos3, cho3[cho]); pos3+=strlen(cho3[cho]);
			strcpy(key3+pos3, jung3[jung]); pos3+=strlen(jung3[jung]);
			strcpy(key3+pos3, jong3[jong]); pos3+=strlen(jong3[jong]);
 
			strcpy(key3f+pos3f, cho3[cho]); pos3f+=strlen(cho3[cho]);
			strcpy(key3f+pos3f, jung3f[jung]); pos3f+=strlen(jung3f[jung]);
			strcpy(key3f+pos3f, jong3f[jong]); pos3f+=strlen(jong3f[jong]);
 
			strcpy(key3n+pos3n, cho3[cho]); pos3n+=strlen(cho3[cho]);
			strcpy(key3n+pos3n, jung3n[jung]); pos3n+=strlen(jung3n[jung]);
			strcpy(key3n+pos3n, jong3n[jong]); pos3n+=strlen(jong3n[jong]);
			i++;
		}
	printf("Hangul characters: %d\n\n",hanguls);
	puts("Result of 2bulsik"); init_type_info(2); parse(key2);
	puts("\nResult of 3bulsik"); init_type_info(3); parse(key3);
	puts("\nResult of 3bulsik final"); init_type_info(3); parse(key3f);
	puts("\nResult of 3bulsik noshift"); init_type_info(3); parse(key3n);
	free(key3); free(key2); free(key3f); free(key3n);
	free(buff);
}
